{-# OPTIONS --no-termination-check #-}

open import lib.Prelude 
open List using (_++_) 
open List.In using (_-_ ; skip)
-- open List.InLists
-- open List.Subsets

open import foc.focbind.univ.examples.DefinedAtoms 
open import foc.focbind.univ.Types
open import foc.focbind.univ.Semantics 

module foc.focbind.univ.Map where

module Map(S : SignatureInterface.Signature) where

  open Semantics(S)
  open SignatureInterface.Signature S

  VarRel = Type -> Vars -> Vars -> Set

  module Compat (Co     : VarRel)
                (Contra : VarRel)
   where
    Compat : Type -> Vars -> Vars -> Set
    Compat (D #) Ψ Ψ' = (D ∈ Ψ) -> (D ∈ Ψ')
    Compat (D⁺ D) Ψ Ψ'   =   ({A : Type} -> (c : InΣ (D ⇐ A)) -> Co A Ψ Ψ') 
                               × ({ch : _} -> ((▹ D {ch}) ∈ Ψ  ->  < Ψ' > D⁺ D))
    Compat 1⁺ Ψ Ψ' = Unit
    Compat 0⁺ Ψ Ψ' = Unit
    Compat (A ⊗ B) Ψ Ψ'    =  Co A Ψ Ψ' × Co B Ψ Ψ'
    Compat (A ⊕ B) Ψ Ψ'    =  Co A Ψ Ψ' × Co B Ψ Ψ'
    Compat (A ⊃ B) Ψ Ψ'    =  Contra A Ψ' Ψ × Co B Ψ Ψ'
    Compat (list A) Ψ Ψ'   =  Co A Ψ Ψ'
    Compat (Ψ0 ⇒* A) Ψ Ψ'  =  Co A (Ψ0 ++ Ψ) (Ψ0 ++ Ψ')
    Compat (□ A) Ψ Ψ'      =  Unit
    Compat (∀c τ) Ψ Ψ'    =  (Ψ0 : Vars) -> Co (τ Ψ0) Ψ Ψ'
    Compat (∀≤ D A) Ψ Ψ'  =  (Ψ0 : Vars) -> AllEq Ψ0 D -> Co A (Ψ0 ++ Ψ) (Ψ0 ++ Ψ')
    Compat (∃c τ) Ψ Ψ'    =  (Ψ0 : Vars) -> Co (τ Ψ0) Ψ Ψ'
    Compat (∀⇒ A) Ψ Ψ'   =  (Ψ0 : Vars) -> Co A (Ψ0 ++ Ψ) (Ψ0 ++ Ψ')
    Compat (∃⇒ A) Ψ Ψ'   =  (Ψ0 : Vars) -> Co A (Ψ0 ++ Ψ) (Ψ0 ++ Ψ')
    Compat ([ Ψ ]* A) _ _ =  Unit

  AllCompat : VarRel -> VarRel -> Set
  AllCompat Co Contra = (A : Type) (Ψ Ψ' : Vars) -> Co A Ψ Ψ' -> Compat.Compat Co Contra A Ψ Ψ'

  AllCompatI : VarRel -> VarRel -> Set
  AllCompatI Co Contra = (A : Type) {Ψ Ψ' : Vars} -> Co A Ψ Ψ' -> Compat.Compat Co Contra A Ψ Ψ'

  AllNotstuck : VarRel -> VarRel -> Set
  AllNotstuck Co Contra = (A : Type) (Ψ Ψ' : Vars) -> Co A Ψ Ψ' 
                        -> Either (< Ψ > A -> < Ψ' > A)  -- if I have such a function, I'm done
                                  (Compat.Compat Co Contra A Ψ Ψ') -- otherwise I can continue

  AllNotstuckI : VarRel -> VarRel -> Set
  AllNotstuckI Co Contra = (A : Type) {Ψ Ψ' : Vars} -> Co A Ψ Ψ' 
                         -> Either (< Ψ > A -> < Ψ' > A)  -- if I have such a function, I'm done
                                   (Compat.Compat Co Contra A Ψ Ψ') -- otherwise I can continue
  
  record MapSpec : Set1 where
   field
    Co      : VarRel
    Contra  : VarRel
    compat  : AllCompat Co Contra

  record MapSpec+ : Set1 where
   field
    Co       : VarRel
    Contra   : VarRel
    notstuck : AllNotstuck Co Contra

  promotespec : MapSpec -> MapSpec+
  promotespec r = record { Co = MapSpec.Co r ; 
                           Contra = MapSpec.Contra r ; 
                           notstuck = \ A Ψ Ψ' co -> Inr (MapSpec.compat r A Ψ Ψ' co) }

  MapType : (Type -> Vars -> Vars -> Set) -> Set
  MapType R = (A : Type) {Ψ Ψ' : Vars} -> (R A Ψ Ψ') -> < Ψ > A -> < Ψ' > A

  module Map1+(S : MapSpec+) 
             -- abstracted over what to do in the contravariant case
             (map' : MapType (MapSpec+.Contra S)) where
    open MapSpec+ S
    open Compat Co Contra
  
    mutual
      map : MapType Co 
      map A {Ψ} {Ψ'} co e with notstuck A Ψ Ψ' co
      ... | Inl done = done e 
      ... | Inr compat = mapc A co compat e 

      mapc : (A : Type) {Ψ Ψ' : Vars} -> (Co A Ψ Ψ') -> (Compat A Ψ Ψ') -> < Ψ > A -> < Ψ' > A
      mapc 1⁺ co compat t = _
      mapc 0⁺ co compat ()
      mapc (A ⊗ B) {Ψ} {Ψ'} co compat (e1 , e2) = map A (fst compat) e1 , 
                                           map B (snd compat) e2 
      mapc (A ⊕ B) {Ψ} {Ψ'} co compat (Inl e1) = Inl (map A (fst compat) e1)
      mapc (A ⊕ B) {Ψ} {Ψ'} co compat (Inr e1) = Inr (map B (snd compat) e1)
      mapc (list A) co compat [] = []
      mapc (list A) {Ψ} {Ψ'} co compat (x :: xs) = map A compat  x :: map (list A) co  xs
      mapc (A ⊃ B) {Ψ} {Ψ'} co compat e = \ y -> (map B (snd compat) (e (map' A (fst compat) y)))
      mapc (Ψ0 ⇒* A) {Ψ} {Ψ'} co compat e = map A compat e
      mapc (∀c τ) {Ψ} {Ψ'} co compat e = \ Ψ0 -> map (τ Ψ0) (compat Ψ0) (e Ψ0) 
      mapc (∀≤ D A) {Ψ} {Ψ'} co compat e = \ Ψ0 ev -> map A (compat Ψ0 ev) (e Ψ0 ev)
      mapc (∃c τ) {Ψ} {Ψ'} co compat (Ψ0 , e) = Ψ0 , map (τ Ψ0) (compat Ψ0) e
      mapc (D⁺ Dat) {Ψ} {Ψ'} co compat (_·_ {A} c e) =  c · map A (fst compat c) e  
      mapc (D⁺ Dat) {Ψ} {Ψ'} co compat (▹ x) = (snd compat x)
      mapc (Dat #) {Ψ} {Ψ'}  co compat x = (compat x)
      mapc (∀⇒ A) {Ψ} {Ψ'}  co compat e = \ Ψ0 -> map A (compat Ψ0) (e Ψ0)
      mapc (∃⇒ A) {Ψ} {Ψ'}  co compat (Ψ0 , e) = Ψ0 , map A (compat Ψ0) e
      mapc (□ A) {Ψ} {Ψ'} co compat e = e
      mapc ([ _ ]* A) {Ψ} {Ψ'} co compat e = e

  module Map1(S : MapSpec) 
             -- abstracted over what to do in the contravariant case
             (map' : MapType (MapSpec.Contra S)) where
    module M = Map1+ (promotespec S) map'

    map : MapType (MapSpec.Co S)
    map A {Ψ} {Ψ'} co e = M.map A co e 

  record MapSpec2 : Set1 where
   field
    s : MapSpec
    compat/contra : AllCompat (MapSpec.Contra s) (MapSpec.Co s)
  
  module MapMutual(S : MapSpec2) where
    open MapSpec (MapSpec2.s S)
  
    -- since it's not going to termination-check anyway, let's use some mutually recursive functor applications
    -- to avoid writing things twice
    mutual 
      map : MapType Co 
      map A i e = Map1.map (record {Co = Co ; Contra = Contra ; compat = compat }) map' A i e
  
      map' : MapType Contra
      map' A i e = Map1.map (record {Co = Contra ; Contra = Co ; compat = MapSpec2.compat/contra S }) map A i e

  module MapSelfDual(Co : VarRel)
                    (compat : AllCompat Co Co) where
  
    map : MapType Co
    map A i e = MapMutual.map (record { s = record {Co = Co ; Contra = Co ; compat = compat} ; compat/contra = compat}) A i e
