
%% simplified version for use in the paper;
%% not actually used in the code

\ignore{
\begin{code}
{-# OPTIONS --no-termination-check #-}

open import lib.Prelude 
open List using (_++_) 
open List.In using (_-_ ; skip)
-- open List.InLists
-- open List.Subsets

open import foc.focbind.univ.examples.DefinedAtoms 
open import foc.focbind.univ.Types
open import foc.focbind.univ.Semantics 

module foc.focbind.univ.MapExpository(S : SignatureInterface.Signature) where

open Semantics(S)
open SignatureInterface.Signature S

VarRel = Type -> Vars -> Vars -> Set

\end{code}}

The structural properties are implemented by instantiating a generic
traversal for \ttt{< Ψ > A}.  The generic traversal has the following
type:

\noagda\begin{code}
 map : (A : Type) {Ψ Ψ' : Vars} 
     -> (Co A Ψ Ψ')  ->  < Ψ > A  ->  < Ψ' > A
\end{code}
%
This should be read as follows: for every \ttt{A Ψ Ψ'}, under the
condition \ttt{Co A Ψ Ψ'}, there is a map from terms of type \ttt{A} in
\ttt{Ψ} to terms of type \ttt{A} in \ttt{Ψ'}.  

\ttt{Co : Type -> Vars -> Vars -> Set} is a \emph{variable relation}, a
type-indexed family of relations between two contexts.  \ttt{Co} is in
fact a parameter to the generic \ttt{map}; it provides: (1) a variable
or term in Ψ' for each variable in Ψ that the traversal runs into; and
(2) enough information to keep the traversal going inductively.  We will
instantiate \ttt{Co} with a specific relation for each traversal; e.g.,
for weakening with a variable of type \ttt{D}, \ttt{Co} will relate
\ttt{Ψ} to \ttt{(Ψ ,, D)} under appropriate conditions on \ttt{D} and
\ttt{A}.

For expository purposes, we present a slightly simplified version of
the traversal first; the generalization is described with weakening below.

\subsection{Compatibility} 

We account for the requirement on \ttt{Co} with the notion of
\emph{compatibility.}  Suppose that \ttt{Co} and \ttt{Contra} are
variable relations.  We say that \ttt{Co} and \ttt{Contra} are
compatible iff there is a term

\noagda\begin{code}
 compat : ({A : Type} {Ψ Ψ' : Vars} 
             -> Co A Ψ Ψ' -> Compat Co Contra A Ψ Ψ')
\end{code}
%
%% (∀ \ttt{A Ψ Ψ'}, if \ttt{Co A Ψ Ψ'} then \ttt{Compat Co Contra A Ψ Ψ'}), 
where \ttt{Compat} is defined as follows:

\ignore{
\begin{code}
module Compat (Co     : VarRel)
              (Contra : VarRel) where
\end{code}}

\begin{code}
 Compat : Type -> Vars -> Vars -> Set
 Compat (D #) Ψ Ψ' = (D ∈ Ψ) -> (D ∈ Ψ')
 Compat (D⁺ D) Ψ Ψ' = 
    ({A : Type} -> (c : InΣ (D ⇐ A)) -> Co A Ψ Ψ') 
  × ({ch : _} -> (▹ D {ch}) ∈ Ψ  ->  < Ψ' > D⁺ D)
 Compat 1⁺ Ψ Ψ' = Unit
 Compat 0⁺ Ψ Ψ' = Unit
 Compat (A ⊗ B) Ψ Ψ' =  Co A Ψ Ψ' × Co B Ψ Ψ'
 Compat (A ⊕ B) Ψ Ψ' =  Co A Ψ Ψ' × Co B Ψ Ψ'
 Compat (A ⊃ B) Ψ Ψ' =  Contra A Ψ' Ψ × Co B Ψ Ψ'
 Compat (list A) Ψ Ψ' =  Co A Ψ Ψ'
 Compat (Ψ0 ⇒* A) Ψ Ψ'  =  Co A (Ψ + Ψ0) (Ψ' + Ψ0)
 Compat (□ A) Ψ Ψ' =  Unit
 Compat (∀c τ) Ψ Ψ' =  (Ψ0 : _) -> Co (τ Ψ0) Ψ Ψ'
 Compat (∀≤ D A) Ψ Ψ' =  
    (Ψ0 : _) -> AllEq Ψ0 D -> Co A (Ψ + Ψ0) (Ψ' + Ψ0)
 Compat (∃c τ) Ψ Ψ'    =  (Ψ0 : _) -> Co (τ Ψ0) Ψ Ψ'
 Compat (∀⇒ A) Ψ Ψ' = 
    (Ψ0 : _) -> Co A (Ψ + Ψ0) (Ψ' + Ψ0)
 Compat (∃⇒ A) Ψ Ψ' = 
    (Ψ0 : _) -> Co A (Ψ + Ψ0) (Ψ' + Ψ0)
 Compat ([ Ψ ]* A) _ _ =  Unit
\end{code}
%

\ttt{Compat} imposes certain conditions on \ttt{Co} and \ttt{Contra}.
For example, for variable types \ttt{D \#}, it says that \ttt{Co (D \#)
Ψ Ψ'} induces a map from variables of type \ttt{D} in Ψ to variables in
\ttt{Ψ'}.  For defined atoms \ttt{D⁺ D}, \ttt{Compat} says that \ttt{Co
(D⁺ D) Ψ Ψ'} induces a map from variables in Ψ to terms\footnote{In
retrospect, \ttt{bind} may have been a better name than \ttt{map},
because \ttt{Co} maps variables to terms, as in monadic
traversals~\citep{altenkirchreus99gadt}.} in \ttt{Ψ'}, and that \ttt{Co
A Ψ Ψ'} holds for every premise of every constant inhabiting \ttt{D}.
In all other cases, \ttt{Compat} provides enough information to keep the
induction going in \ttt{map} below.  This amounts to insisting that
\ttt{Co} (or \ttt{Contra}) holds on the subexpressions of a type in all
appropriate contexts.  For example, the condition for \ttt{Ψ0 ⇒* A} is
that \ttt{Co} holds for \ttt{A} in the contexts extended with \ttt{Ψ0}.
As discussed in Section~\ref{sec:defs}, ∀⇒ and ∃⇒ and \ttt{[ \_ ]*} are
actually built-in types, so they can be handled specially by the
structural property tactics, though their semantics is the same as if
they were derived forms.  

In the usual monadic traversals of syntax~\citep{altenkirchreus99gadt},
\ttt{Co \_ Ψ Ψ'} is taken to be \ttt{(D : VarType) -> D ∈ Ψ -> < Ψ' >
D}---i.e.  a realization of every variable in Ψ as a term in Ψ'.  In our
setting, this does not suffice to define a traversal, because (1) it
does not provide for the contravariant flip necessary to process the
domains of computational functions and (2) it does not allow us to
express a \emph{conditional} traversal, where conditions on the types
ensure that the traversal will only find certain variables, and thus
that only those variables need realizations.  Compatibility ensures that
\ttt{Co} provides enough information \ttt{Contra} to process the
contravariant positions to the left of a computational arrow.
Additionally, it permits conditional traversals: below, we will
instantiate \ttt{Co} so that it is uninhabited for certain \ttt{A}.

\ignore{
\begin{code}
AllCompat : VarRel -> VarRel -> Set
AllCompat Co Contra = {A : Type} {Ψ Ψ' : Vars} -> Co A Ψ Ψ' -> Compat.Compat Co Contra A Ψ Ψ'

record MapSpec : Set1 where
 field
  Co      : VarRel
  Contra  : VarRel
  compat  : AllCompat Co Contra

module Map1(S : MapSpec) 
           -- abstracted over what to do in the contravariant case
           (map' : (A : Type) {Ψ Ψ' : Vars} -> ((MapSpec.Contra S) A Ψ Ψ') -> < Ψ > A -> < Ψ' > A) where
 open MapSpec S
 open Compat Co Contra
\end{code}}

\subsection{Map}

Suppose that \ttt{Co} and \ttt{Contra} are compatible, and assume a function 

\noagda\begin{code}
 map' : (A : Type) {Ψ Ψ' : Vars} 
      -> (Contra A Ψ Ψ') -> < Ψ > A -> < Ψ' > A
\end{code}
%
that is the equivalent of \ttt{map} for the \ttt{Contra}variant positions.

Then we implement \ttt{map} in Figure~\ref{fig:map}.  In the first and
second cases, the compatibility of \ttt{Co} induces the map on variables
that we need.  In the third case, we pre-compose the function with
\ttt{map'} and post-compose with \ttt{map}.  In all other cases,
\ttt{map} simply commutes with constructors, or stops early if it hits a
boxed term.

\begin{figure}
\begin{code}
 map : (A : Type) {Ψ Ψ' : Vars} 
     -> (Co A Ψ Ψ') -> < Ψ > A -> < Ψ' > A
 map (D⁺ Dat) co (▹ x) = (snd (compat co) x)
 map (Dat #)  co x = ((compat co) x)
 map (A ⊃ B) co e = 
   \ y -> (map B (snd (compat co)) 
               (e (map' A (fst (compat co)) y)))
 map 1⁺ co t = _
 map 0⁺ co ()
 map (A ⊗ B) {Ψ} {Ψ'} co  (e1 , e2) = 
   (map A (fst (compat co)) e1 , 
    map B (snd (compat co)) e2)
 map (A ⊕ B) co (Inl e1) = 
   Inl (map A (fst (compat co)) e1)
 map (A ⊕ B) co (Inr e1) = 
   Inr (map B (snd (compat co)) e1)
 map (list A) co [] = []
 map (list A) co (x :: xs) = 
   map A (compat co) x :: map (list A) co xs
 map (Ψ0 ⇒* A) co e = map A (compat co) e
 map (∀c τ)    co e = 
   \ Ψ0 -> map (τ Ψ0) ((compat co) Ψ0) (e Ψ0) 
 map (∀≤ D A)  co e = 
   \ Ψ0 ev -> map A ((compat co) Ψ0 ev) (e Ψ0 ev)
 map (∃c τ)    co (Ψ0 , e) = 
   Ψ0 , map (τ Ψ0) ((compat co) Ψ0) e
 map (D⁺ Dat)  co (_·_ {A} c e) = 
   c · map A (fst (compat co) c) e  
 map (∀⇒ A)   co e = 
   \ Ψ0 -> map A ((compat co) Ψ0) (e Ψ0)
 map (∃⇒ A)   co (Ψ0 , e) = 
   Ψ0 , map A ((compat co) Ψ0) e
 map (□ A)     co e = e
 map ([ Ψ ]* A) co e = e
\end{code}
\caption{Map}
\label{fig:map}
\end{figure}