\ignore{
\begin{code}
{-# OPTIONS --no-termination-check #-}

open import lib.Prelude 
open List using (_++_) 
open List.In using (_-_ ; skip)
-- open List.InLists
-- open List.Subsets

open import foc.focbind.univ.examples.DefinedAtoms 
open import foc.focbind.univ.Types 
open import foc.focbind.univ.Semantics 
open import foc.focbind.univ.Map

module foc.focbind.univ.StrengthenT where

module Strengthen(S : SignatureInterface.Signature) where
  open Semantics(S)
  open SignatureInterface.Signature S
  open Map S

  mutual
    -- FIXME: convince of termination
    --        might be more efficient to share the visited list across recursive calls
    irrel* : List DefAtom -> DefAtom -> Type -> Bool
    irrel* seen D 1⁺ = True
    irrel* seen D 0⁺ = True
    irrel* seen D (A ⊗ B)  = (irrel* seen D A) andalso (irrel* seen D B)
    irrel* seen D (A ⊕ B) = irrel* seen D A andalso irrel* seen D B
    irrel* seen D (list A)  = irrel* seen D A
    irrel* seen D (D⁺ Dat) with List.In.find-in (\D0 -> DefinedAtoms.eq D0 Dat) seen
    ...                           | Some _ = True
    ...                           | None   = irrelD (Dat :: seen) D Dat
    irrel* seen D (Dat #)  = Sums.not (DefinedAtoms.eq D (un▹ Dat))
    irrel* seen D (Ψ ⇒* A) = irrel* seen D A
    irrel* seen D (□ A)    = True
    irrel* seen D (A ⊃ B)  = irrel* seen D A andalso irrel* seen D B
    irrel* seen D (∃⇒ A) = irrel* seen D A
    irrel* seen D (∀⇒ A) = irrel* seen D A
    irrel* seen D (∀≤ _ A) = irrel* seen D A
    irrel* seen D (∃c _) = False -- punt
    irrel* seen D (∀c _) = False
    irrel* seen D ([ Ψ ]* A)    = True

    irrelD : List DefAtom -> DefAtom -> DefAtom -> Bool
    irrelD seen D Dat = Sums.not (DefinedAtoms.eq D Dat) andalso 
                        List.all (\ A -> irrel* seen D A) (premises Dat)

  unstick : ∀ {seen D Dat} 
          -> Check(irrel* seen D (D⁺ Dat)) 
          -> Either (Check(irrelD (Dat :: seen) D Dat)) (Σ \(D0 : DefAtom) -> (D0 ∈ seen) × Check(DefinedAtoms.eq D0 Dat))
  unstick {seen} {D} {Dat} ch with List.In.find-in (\D0 -> DefinedAtoms.eq D0 Dat) seen
  ...                           | Some p = Inr p 
  ...                           | None = Inl ch 

  irrel : DefAtom -> Type -> Bool
  irrel = irrel* [] 

  open List.Subset
  
  mutual
    -- FIXME: termination
    -- computational induction on irrel* should work
    irrel*-mono : {Ds : List DefAtom} {Ds' : List DefAtom} -> (Ds ⊆ Ds')
                -> {D : DefAtom} -> (A : Type) -> Check (irrel* Ds D A) -> Check (irrel* Ds' D A)
    irrel*-mono w 1⁺ c = _
    irrel*-mono w 0⁺ _ = _
    irrel*-mono {seen} w {D} (A ⊗ B) c = 
      check-andI (irrel*-mono w A (fst (check-andE c))) 
                 (irrel*-mono w B (snd (check-andE{ irrel* seen D A } c))) 
    irrel*-mono {seen} w {D} (A ⊕ B) c = 
      check-andI (irrel*-mono w A (fst (check-andE c))) 
                 (irrel*-mono w B (snd (check-andE{ irrel* seen D A } c))) 
    irrel*-mono w (list A) c = irrel*-mono w A c
    irrel*-mono w (Ψ ⇒* A) c = irrel*-mono w A c
    irrel*-mono w (□ A) c = c
    irrel*-mono {seen} w {D} (A ⊃ B) c = 
      check-andI (irrel*-mono w A (fst (check-andE c))) 
                 (irrel*-mono w B (snd (check-andE{ irrel* seen D A } c))) 
    irrel*-mono w (∃c _) () 
    irrel*-mono w (∀c _) () 
    irrel*-mono w (Dat #) c = c 
    irrel*-mono {seen} {seen'} w (D⁺ Dat) c with List.In.find-in (\D0 -> DefinedAtoms.eq D0 Dat) seen 
    ...                         | None with List.In.find-in (\D0 -> DefinedAtoms.eq D0 Dat) seen' 
    ...                                   | Some _ = _ 
    ...                                   | None = irrelD-mono{Dat :: seen}{Dat :: seen'} (⊆-::-cong w) Dat c  
    irrel*-mono {seen} {seen'} w (D⁺ Dat) c | Some (_ , i , p) 
      with List.In.find-in (\D0 -> DefinedAtoms.eq D0 Dat) seen' | List.In.must-find (\D0 -> DefinedAtoms.eq D0 Dat) seen' (w i) p 
    ... | .(Some (x' , i' , p')) | (x' , i' , p' , Refl) = _ 
    irrel*-mono w (∃⇒ A) c = irrel*-mono w A c
    irrel*-mono w (∀⇒ A) c = irrel*-mono w A c
    irrel*-mono w (∀≤ D A) c = irrel*-mono w A c
    irrel*-mono w ([ _ ]* A) c = _

    irrelD-mono : {Ds : List DefAtom} {Ds' : List DefAtom} -> (Ds ⊆ Ds')
                -> {D : DefAtom} -> (Dat : DefAtom) -> Check (irrelD Ds D Dat) -> Check (irrelD Ds' D Dat)
    irrelD-mono {Ds} {Ds'} w {D} Dat c = check-andI{Sums.not (DefinedAtoms.eq D Dat)}{_} 
                                        (fst (check-andE c))
                                        newall where 
     eworig = List.EW.from-checkall (premises Dat) (snd (check-andE{Sums.not (DefinedAtoms.eq D Dat)} c))
     ewnew  = List.EW.mapew (\ {A} c -> irrel*-mono w  A  c  ) eworig 
     newall = List.EW.to-checkall {Type} {(premises Dat)} { _ } ewnew  

  StrFn : (A : Type) -> (D : VarType) -> Set
  StrFn A D = ∀ {Ψ} -> (i : D ∈ Ψ) -> < Ψ > A -> < Ψ - i > A

  WknFn : (A : Type) -> (D : VarType) -> Set
  WknFn A D = ∀ {Ψ} -> (i : D ∈ Ψ) -> < Ψ - i > A -> < Ψ > A 

  Seen : DefAtom -> Set
  Seen D = Σ \ (seen : List DefAtom) -> ({Dat : DefAtom} -> (Dat ∈ seen) -> Check(irrelD seen D Dat))

  Co : Type -> Vars -> Vars -> Set
  Co A Ψ Ψ' = Σ \(D : VarType) -> 
                 Σ \(i : D ∈ Ψ) -> 
                   Σ \(seen : Seen (un▹ D)) -> Check(irrel* (fst seen) (un▹ D) A) × Id Ψ' (Ψ - i)

  Contra : Type -> Vars -> Vars -> Set
  Contra A Ψ Ψ' = Co A Ψ' Ψ 

  okvar : {D : _} {Ψ : Vars} (Dat : DefAtom)
        -> (i : D ∈ Ψ)
        -> Check(Sums.not (DefinedAtoms.eq (un▹ D) Dat))
        -> ({ch' : _} -> (▹ Dat {ch'}) ∈ Ψ -> (▹ Dat {ch'}) ∈ (Ψ - i))
  okvar {D} Dat i ch {ch'} j with List.In.indeq i j 
  okvar {.(▹ Dat {ch'})} Dat i ch {ch'} j | Inl Refl = Sums.abort (check-noncontra(DefinedAtoms.eq Dat Dat) (DefinedAtoms.eq-refl Dat ) ch) 
  ... | Inr newj = newj

  compat : AllCompatI Co Contra
  compat ((▹ Dat {c}) #) {Ψ} (D , i , seen , ch , Refl) = okvar Dat i ch 
  compat (D⁺ Dat) (D , i , seen , ch , Refl) with List.In.find-in (\D0 -> DefinedAtoms.eq D0 Dat) (fst seen)
  ... | Some (Dat' , DatInSeeen , cheq) = 
   (\ {_} -> f) , 
   \ x -> ▹ (okvar Dat i 
                (fst (check-andE (snd seen ((Id.subst (\ x -> x ∈ (fst seen)) (DefinedAtoms.certify-eq cheq) DatInSeeen ) ))) ) 
                x)    where
    f : {A : Type} (c : InΣ (Dat ⇐ A)) -> Co A _ _
    f {A} c = D , i , seen , (List.In.in-all (\ A -> irrel* (fst seen) (un▹ D ) A)
                                       (snd (check-andE{Sums.not (DefinedAtoms.eq (un▹ D) Dat)} 
                                                       ( (snd seen) ((Id.subst (\ x -> x ∈ (fst seen)) (DefinedAtoms.certify-eq cheq) DatInSeeen) ) )))
                                       {A} 
                                       (certify-premises c))  , Refl 
  ... | None = (\ {_} -> f) , \ x -> ▹(okvar Dat i (fst (check-andE{(Sums.not (DefinedAtoms.eq (un▹ D) Dat))}{_} ch)) x) where
    f : {A : Type} (c : InΣ (Dat ⇐ A)) -> Co A _ _
    f {A} c = (D , i , 
               (Dat :: (fst seen) , 
                 \ {Dat'} i -> List.In.::-every {DefAtom}{Dat}
                              {\Dat'' -> Check(irrelD (Dat :: (fst seen)) (un▹ D) Dat'')}
                              {fst seen} 
                              (\ {Dat''} x ->   irrelD-mono iS  Dat'' ((snd seen) x)     ) ch  {Dat'} i ) , 
               (List.In.in-all (\ A -> irrel* (Dat :: (fst seen)) (un▹ D ) A)  
                         (snd (check-andE{Sums.not (DefinedAtoms.eq (un▹ D) Dat)} ch)) {A} ((certify-premises c)))  , 
               Refl)
  compat 1⁺ _ = _ 
  compat 0⁺ _ = _ 
  compat (A ⊗ B) (D , i , seen , ch , Refl) = ((D , i , seen ,  (fst (check-andE ch))  , Refl) ) ,
        (D , i , seen , (snd (check-andE{(irrel* (fst seen) (un▹ D) A)} ch)) , Refl)
  compat (A ⊕ B) (D , i , seen , ch , Refl) = ((D , i , seen ,  (fst (check-andE ch))  , Refl) ) ,
        (D , i , seen , (snd (check-andE{(irrel* (fst seen) (un▹ D) A)} ch)) , Refl)
  compat (□ A) _ = _ 
  compat (list A) inv = inv
  compat (∀c _) (_ , _ , _ , () , _)
  compat (∃c _) (_ , _ , _ , () , _)
  compat (Ψ ⇒* A) (D , i , seen , ch , Refl) = (D , skip Ψ i , seen  , ch  , (List.In.remove/append Ψ i) )
  compat (A ⊃ B) (D , i , seen , ch , Refl) = 
    (D , i , seen , (fst (check-andE{(irrel* (fst seen) (un▹ D) A)} ch)) , Refl), 
    (D , i , seen , (snd (check-andE{(irrel* (fst seen) (un▹ D) A)} ch)) , Refl)
  compat (∀⇒ A) (D , i , seen , ch , Refl) = \ Ψ -> (D , skip Ψ i , seen  , ch  , (List.In.remove/append Ψ i) )
  compat (∀≤ Bound A) (D , i , seen , ch , Refl) = \ Ψ ev -> (D , skip Ψ i , seen  , ch  , (List.In.remove/append Ψ i) )
  compat (∃⇒ A) (D , i , seen , ch , Refl) = \ Ψ -> (D , skip Ψ i , seen  , ch  , (List.In.remove/append Ψ i) )
  compat ([ _ ]* A) _ = _ 

  -- FIXME: abstract!
  compat/contra : AllCompatI Contra Co
  compat/contra ((▹ Dat {c}) #) (D , i , seen , ch , Refl) = \ x -> (List.Subset.⊆remove i x )
  compat/contra (D⁺ Dat) (D , i , seen , ch , Refl) with List.In.find-in (\D0 -> DefinedAtoms.eq D0 Dat) (fst seen)
  ... | Some (Dat' , DatInSeeen , cheq) = 
   (\ {_} -> f) , 
   \ x -> ▹ (List.Subset.⊆remove i x )    where
    f : {A : Type} (c : InΣ (Dat ⇐ A)) -> Contra A _ _
    f {A} c = D , i , seen , (List.In.in-all (\ A -> irrel* (fst seen) (un▹ D ) A)
                                       (snd (check-andE{Sums.not (DefinedAtoms.eq (un▹ D) Dat)} 
                                                       ( (snd seen) ((Id.subst (\ x -> x ∈ (fst seen)) (DefinedAtoms.certify-eq cheq) DatInSeeen) ) )))
                                       {A} 
                                       (certify-premises c))  , Refl 
  ... | None = (\ {_} -> f) , \ x -> ▹ (List.Subset.⊆remove i x )  where
    f : {A : Type} (c : InΣ (Dat ⇐ A)) -> Contra A _ _
    f {A} c = (D , i , 
               (Dat :: (fst seen) , 
                 \ {Dat'} i -> List.In.::-every {DefAtom}{Dat}
                              {\Dat'' -> Check(irrelD (Dat :: (fst seen)) (un▹ D) Dat'')}
                              {fst seen} 
                              (\ {Dat''} x ->   irrelD-mono iS  Dat'' ((snd seen) x)     ) ch  {Dat'} i ) , 
               (List.In.in-all (\ A -> irrel* (Dat :: (fst seen)) (un▹ D ) A)  
                         (snd (check-andE{Sums.not (DefinedAtoms.eq (un▹ D) Dat)} ch)) {A} ((certify-premises c)))  , 
               Refl)
  compat/contra 1⁺ _ = _ 
  compat/contra 0⁺ _ = _ 
  compat/contra (A ⊗ B) (D , i , seen , ch , Refl) = ((D , i , seen ,  (fst (check-andE ch))  , Refl) ) ,
        (D , i , seen , (snd (check-andE{(irrel* (fst seen) (un▹ D) A)} ch)) , Refl)
  compat/contra (A ⊕ B) (D , i , seen , ch , Refl) = ((D , i , seen ,  (fst (check-andE ch))  , Refl) ) ,
        (D , i , seen , (snd (check-andE{(irrel* (fst seen) (un▹ D) A)} ch)) , Refl)
  compat/contra (□ A) _ = _ 
  compat/contra (list A) inv = inv
  compat/contra (∀c _) (_ , _ , _ , () , _)
  compat/contra (∃c _) (_ , _ , _ , () , _)
  compat/contra (Ψ ⇒* A) (D , i , seen , ch , Refl) = (D , skip Ψ i , seen  , ch  , (List.In.remove/append Ψ i) )
  compat/contra (A ⊃ B) (D , i , seen , ch , Refl) = 
    (D , i , seen , (fst (check-andE{(irrel* (fst seen) (un▹ D) A)} ch)) , Refl), 
    (D , i , seen , (snd (check-andE{(irrel* (fst seen) (un▹ D) A)} ch)) , Refl)
  compat/contra (∀⇒ A) (D , i , seen , ch , Refl) = \ Ψ -> (D , skip Ψ i , seen  , ch  , (List.In.remove/append Ψ i) )
  compat/contra (∃⇒ A) (D , i , seen , ch , Refl) = \ Ψ -> (D , skip Ψ i , seen  , ch  , (List.In.remove/append Ψ i) )
  compat/contra ([ _ ]* A) _ = _ 
  compat/contra (∀≤ Bound A) (D , i , seen , ch , Refl) = \ Ψ ev -> (D , skip Ψ i , seen  , ch  , (List.In.remove/append Ψ i) )

  Spec : MapSpec2
  Spec = record { s = record {Co = Co ; Contra = Contra ; compat = \ A Ψ Ψ' -> compat A} ; compat/contra = \ A Ψ Ψ' -> compat/contra A  } 

  strengthen-map : MapType Co
  strengthen-map = MapMutual.map Spec

  weaken-map : MapType Contra
  weaken-map = MapMutual.map' Spec

  canStrengthen = irrel

  strengthen/anywhere : (A : Type) {D : VarType} -> {_ : Check(irrel (un▹ D) A)} -> StrFn A D
  strengthen/anywhere A {D} {ch} i e = strengthen-map A ((D , i , ([] , \ ()) , ch , Refl)) e

  strengthen : (A : Type) {D : VarType} {_ : Check (canStrengthen (un▹ D) A)} -> ∀⇛ (D ⇒ A) ⊃ A
  strengthen A {D} {ch} Ψ t = strengthen/anywhere A {_} {ch} i0 t 

  canWeaken/irrel = irrel

  weaken/irrel/anywhere : (A : Type) {D : VarType} -> {_ : Check(irrel (un▹ D) A)} -> WknFn A D
  weaken/irrel/anywhere A {D} {ch} i e = MapMutual.map' Spec A ((D , i , ([] , \ ()) , ch , Refl)) e

  weaken/irrel : (A : Type) {D : VarType} {_ : Check (canWeaken/irrel (un▹ D) A)} -> ∀⇛ A ⊃ (D ⇒ A)
  weaken/irrel A {D} {ch} Ψ t = weaken/irrel/anywhere A {_} {ch} i0 t 

\end{code}}