{-# OPTIONS --no-termination-check #-}

open import lib.Prelude 
open List using (_++_) 
open List.In using (_-_ ; skip)
-- open List.InLists
-- open List.Subsets

open import foc.focbind.univ.examples.DefinedAtoms 
open import foc.focbind.univ.Types 
open import foc.focbind.univ.Semantics 
open import foc.focbind.univ.StrengthenT
open import foc.focbind.univ.WeakenT
open import foc.focbind.univ.Map

module foc.focbind.univ.Subst where

module Subst(S : SignatureInterface.Signature) where
  open Semantics(S)
  module Str = Strengthen(S)
  module Wkn =  Weaken(S)
  open Map S
  open SignatureInterface.Signature S

  mutual
    -- FIXME: convince of termination
    --        might be more efficient to share the visited list across recursive calls

    -- checks two things:
    -- (1) no (Cut #) [ this is a side condition insisting that it's a variable ]
    -- (2) Cut can be weakened with all of the binders
    -- substitution also requires (canWeaken Cut A), which checks for negative occurrences
    -- FIXME: just call irrel on the LHS of ⊃ here;
    --        that will simplify the proof below
    canSubst2* : List DefAtom -> DefAtom -> Type -> Bool
    canSubst2* seen Cut 1⁺ = True
    canSubst2* seen Cut 0⁺ = True
    canSubst2* seen Cut (□ A)  = True
    canSubst2* seen Cut ([ _ ]* A)  = True
    canSubst2* seen Cut (A ⊗ B)  = (canSubst2* seen Cut A) andalso (canSubst2* seen Cut B)
    canSubst2* seen Cut (A ⊕ B)  = (canSubst2* seen Cut A) andalso (canSubst2* seen Cut B)
    canSubst2* seen Cut (list A) = canSubst2* seen Cut A
    canSubst2* seen Cut (D⁺ Dat) with List.In.find-in (\D0 -> DefinedAtoms.eq D0 Dat) seen
    ...                           | Some _ = True -- FIXME: seems to work
    ...                           | None   = canSubst2D (Dat :: seen) Cut Dat
    canSubst2* seen Cut (Dat #) = Sums.not (DefinedAtoms.eq Cut (un▹ Dat))
    canSubst2* seen Cut (Ψ ⇒* A) = (List.all (\ V -> Wkn.canWeaken (un▹ V) (D⁺ Cut)) Ψ) andalso (canSubst2* seen Cut A)
    -- really means "binds before box"
    canSubst2* seen Cut (A ⊃ B)  = (canSubst2* seen Cut A) andalso (canSubst2* seen Cut B)
    canSubst2* seen Cut (∃c _)  = False -- punt: can't tell what it binds
    canSubst2* seen Cut (∀c _)  = False
    canSubst2* seen Cut (∃⇒ _)  = False
    canSubst2* seen Cut (∀⇒ _)  = False
    canSubst2* seen Cut (∀≤ _ _)  = False

    canSubst2D : List DefAtom -> DefAtom -> DefAtom -> Bool
    canSubst2D seen Cut Dat = List.all (\ A -> canSubst2* seen Cut A) (premises Dat)

  canSubst2 : DefAtom -> Type -> Bool
  canSubst2 = canSubst2* []  

  canSubst : DefAtom -> Type -> Bool
  canSubst Cut A = (Wkn.canWeaken Cut A) andalso (canSubst2 Cut A)

  open List.Subset

  mutual
    -- FIXME: termination
    -- computational induction on canSubst2* should work
    canSubst2*-mono : {Ds : List DefAtom} {Ds' : List DefAtom} -> (Ds ⊆ Ds')
                -> {D : DefAtom} -> (A : Type) -> Check (canSubst2* Ds D A) -> Check (canSubst2* Ds' D A)
    canSubst2*-mono w 1⁺ c = _
    canSubst2*-mono w 0⁺ _ = _
    canSubst2*-mono {seen} w {D} (A ⊗ B) c = 
      check-andI (canSubst2*-mono w A (fst (check-andE c))) 
                 (canSubst2*-mono w B (snd (check-andE{ canSubst2* seen D A } c))) 
    canSubst2*-mono {seen} w {D} (A ⊕ B) c = 
      check-andI (canSubst2*-mono w A (fst (check-andE c))) 
                 (canSubst2*-mono w B (snd (check-andE{ canSubst2* seen D A } c))) 
    canSubst2*-mono w (list A) c = canSubst2*-mono w A c
    canSubst2*-mono w {D} (Ψ ⇒* A) c = check-andI (fst (check-andE{(List.all (\ V -> Wkn.canWeaken (un▹ V) (D⁺ D)) Ψ)} c) )
                                                   (canSubst2*-mono w A ((snd (check-andE{(List.all (\ V -> Wkn.canWeaken (un▹ V) (D⁺ D)) Ψ)} c) ) ))
    canSubst2*-mono w (□ A) c = c
    canSubst2*-mono {seen} w {D} (A ⊃ B) c = 
       check-andI (canSubst2*-mono w A (fst (check-andE{(canSubst2* seen D A)} c)))
                   (canSubst2*-mono w B (snd (check-andE{(canSubst2* seen D A)} c))) 
    canSubst2*-mono w (∃c _) () 
    canSubst2*-mono w (∀c _) () 
    canSubst2*-mono w (∃⇒ _) () 
    canSubst2*-mono w (∀⇒ _) () 
    canSubst2*-mono w (∀≤ _ _) () 
    canSubst2*-mono w (Dat #) c = c 
    canSubst2*-mono {seen} {seen'} w (D⁺ Dat) c with List.In.find-in (\D0 -> DefinedAtoms.eq D0 Dat) seen 
    ...                         | None with List.In.find-in (\D0 -> DefinedAtoms.eq D0 Dat) seen' 
    ...                                   | Some _ = _ 
    ...                                   | None =  canSubst2D-mono{Dat :: seen}{Dat :: seen'} (⊆-::-cong w) Dat c  
    canSubst2*-mono {seen} {seen'} w (D⁺ Dat) c | Some (_ , i , p) 
      with List.In.find-in (\D0 -> DefinedAtoms.eq D0 Dat) seen' | List.In.must-find (\D0 -> DefinedAtoms.eq D0 Dat) seen' (w i) p 
    ... | .(Some (x' , i' , p')) | (x' , i' , p' , Refl) = _ 
    canSubst2*-mono w ([ _ ]* _) _ = _

    canSubst2D-mono : {Ds : List DefAtom} {Ds' : List DefAtom} -> (Ds ⊆ Ds')
                -> {D : DefAtom} -> (Dat : DefAtom) -> Check (canSubst2D Ds D Dat) -> Check (canSubst2D Ds' D Dat)
    canSubst2D-mono {Ds} {Ds'} w {D} Dat c =  newall  where
      eworig = List.EW.from-checkall (premises Dat) c
      ewnew  = List.EW.mapew (\{A} c -> canSubst2*-mono w  A  c  ) eworig 
      newall = List.EW.to-checkall{Type}{(premises Dat)}{_} ewnew  


  InvD : (Cut : DefinedAtoms.Atom) -> (A : _) -> (seen : List DefAtom) -> Set
  InvD Cut Dat seen = Check(Wkn.canWeakenD seen Cut Dat) × Check(canSubst2D seen Cut Dat)

  Inv : (Cut : DefinedAtoms.Atom) -> (A : _) -> (seen : List DefAtom) -> Set
  Inv Cut A seen = Check(Wkn.canWeaken* seen Cut A) × Check(canSubst2* seen Cut A)

  Seen : DefAtom -> Set
  Seen Cut = Σ (\ (seen : List DefAtom) ->  ({Dat : DefAtom} -> (Dat ∈ seen) -> InvD Cut Dat seen) )

  data SubstCo (A : Type) (Ψ : Vars) : (Ψ' : Vars) -> Set where
     MakeCo : {Cut : VarType} 
              (for    : Cut ∈ Ψ) 
              (plugin : < Ψ - for > (Cut ⁺)) 
              (seen   : Seen (un▹ Cut)) 
              -> Inv (un▹ Cut) A (fst seen) 
              -> SubstCo A Ψ (Ψ - for)

  Contra = Str.Contra

  postulate 
    V : Void

  subvars : {Ψ : _} {Dat : _} {Cut : _} {ch : Check (DefinedAtoms.world Dat)}
          -> (i : (▹ Dat {ch}) ∈ Ψ)
          -> (for : Cut ∈ Ψ)
          -> < Ψ - for > (Cut ⁺)
          -> < Ψ - for > (D⁺ Dat)
  subvars i for subin with List.In.indeq for i 
  subvars i for subin | Inr newj = ▹ newj  
  ...                 | Inl Refl = subin 

  compat : AllCompatI SubstCo Contra
  compat 1⁺ _ = _ 
  compat 0⁺ _ = _ 
  compat (□ A) _ = _ 
  compat (∀c A) (MakeCo for p seen (_ , ()))
  compat (∃c A) (MakeCo for p seen (_ , ()))
  compat (∀⇒ A) (MakeCo for p seen (_ , ()))
  compat (∀≤ _ A) (MakeCo for p seen (_ , ()))
  compat (∃⇒ A) (MakeCo for p seen (_ , ()))
  compat (A ⊃ B) (MakeCo {Cut} for p seen (wknA , cs)) = 
    (Cut , for , ([] , \()) , fst (check-andE wknA)  , Refl ) , 
    MakeCo for p seen ( snd (check-andE{Str.irrel (un▹ Cut) A} wknA)  ,  snd (check-andE{canSubst2* (fst seen) (un▹ Cut) A} cs)     )    
  compat (A ⊗ B) (MakeCo {Cut} for p seen (wknA , cs)) = 
    MakeCo for p seen ( fst (check-andE{Wkn.canWeaken* (fst seen) (un▹ Cut) A} wknA)  ,  fst (check-andE{canSubst2* (fst seen) (un▹ Cut) A} cs)     )    ,
    MakeCo for p seen ( snd (check-andE{Wkn.canWeaken* (fst seen) (un▹ Cut) A} wknA)  ,  snd (check-andE{canSubst2* (fst seen) (un▹ Cut) A} cs)     )    
  compat (A ⊕ B) (MakeCo {Cut} for p seen (wknA , cs)) = 
    MakeCo for p seen ( fst (check-andE{Wkn.canWeaken* (fst seen) (un▹ Cut) A} wknA)  ,  fst (check-andE{canSubst2* (fst seen) (un▹ Cut) A} cs)     )    ,
    MakeCo for p seen ( snd (check-andE{Wkn.canWeaken* (fst seen) (un▹ Cut) A} wknA)  ,  snd (check-andE{canSubst2* (fst seen) (un▹ Cut) A} cs)     )    
  compat (list A) (MakeCo {Cut} for p seen (wknA , cs)) = MakeCo for p seen ( wknA  ,  cs )
  compat ((▹ Dat) #) (MakeCo {Cut} for p seen (wkn , cs)) = \ i -> Str.okvar{Cut} Dat for  cs i 
  compat (D⁺ Dat) (MakeCo {Cut} for p seen (wkn , cs)) with List.In.find-in (\D0 -> DefinedAtoms.eq D0 Dat) (fst seen)
  ... | None   = 
      (\ {A} c -> let
                     chw = (List.In.in-all (\ A -> Wkn.canWeaken* (Dat :: (fst seen)) (un▹ Cut) A)  
                                            wkn {A} ((certify-premises c))) 
                     chs = (List.In.in-all (\ A -> canSubst2* (Dat :: (fst seen)) (un▹ Cut) A)  
                                            cs {A} ((certify-premises c))) 
                  in MakeCo for p 
                            (Dat :: fst seen , 
                             \ {Dat'} inseen -> (List.In.::-every {DefAtom}{Dat}
                                                 {\Dat'' -> Check(Wkn.canWeakenD (Dat :: (fst seen)) (un▹ Cut) Dat'')}
                                                 {fst seen} 
                                                 (\ {Dat''} x ->   Wkn.canWeakenD-mono iS  Dat'' (fst ((snd seen) x))     ) wkn  {Dat'} inseen)  , 
                                                (List.In.::-every {DefAtom}{Dat}
                                                 {\Dat'' -> Check(canSubst2D (Dat :: (fst seen)) (un▹ Cut) Dat'')}
                                                 {fst seen} 
                                                 (\ {Dat''} x ->   canSubst2D-mono iS  Dat'' (snd ((snd seen) x))     ) cs  {Dat'} inseen)   ) 
                            (chw , chs) ) , 
      \ i -> subvars i for p  
  ... | Some (Dat' , Dat'InSeen , checkeq) =
      (\ {A} c -> 
       let
        chw : Check (Wkn.canWeaken* (fst seen) (un▹ Cut) A)
        chw = (List.In.in-all (\ A -> Wkn.canWeaken* (fst seen) (un▹ Cut ) A)
                              ((fst ((snd seen)
                                ((Id.subst (\ x -> x ∈ (fst seen)) 
                                           (DefinedAtoms.certify-eq checkeq) Dat'InSeen) ) )) )
                              {A} 
                              (certify-premises c))
        chs = (List.In.in-all (\ A -> canSubst2* (fst seen) (un▹ Cut ) A)
                              (snd ((snd seen) ((Id.subst (\ x -> x ∈ (fst seen)) 
                                           (DefinedAtoms.certify-eq checkeq) Dat'InSeen) )))
                              {A} 
                              (certify-premises c))
       in 
         MakeCo for p seen ( chw  , chs ) 
      ) , 
      \ i -> subvars i for p   
  compat (Ψ ⇒* A) {Ψ1} (MakeCo {▹ Cut {ch}} for p seen (wkn , cs)) = 
    Id.subst (\ x -> SubstCo A (Ψ ++ Ψ1) x) (Id.sym (List.In.remove/append Ψ for)) 
             (MakeCo (skip Ψ for) 
                     (Id.subst (\ x -> < x > (D⁺ Cut)) 
                               (List.In.remove/append Ψ for) 
                               (Wkn.weaken* (D⁺ Cut) Ψ {(fst (check-andE{(List.all (\ V -> Wkn.canWeaken (un▹ V) (D⁺ Cut)) Ψ)} cs))} _ p )  )  
                     seen 
                     ( wkn  , 
                       snd (check-andE{(List.all (\ V -> Wkn.canWeaken (un▹ V) (D⁺ Cut)) Ψ)} cs)  ))  
  compat ([ _ ]* _) _ = _

  Spec = record { Co = SubstCo ; Contra = Contra ; compat = \ A Ψ Ψ' -> compat A } 

  subst/anywhere : (A : Type) (Cut : VarType) 
         -> {_ : Check(Wkn.canWeaken (un▹ Cut) A)} {_ : Check(canSubst2 (un▹ Cut) A)}
         -> ∀ {Ψ} -> (i : Cut ∈ Ψ) -> < Ψ - i > (Cut ⁺) -> < Ψ > A -> < Ψ - i > A
  subst/anywhere A D {ch1} {ch2} i arg body = Map1.map Spec Str.weaken-map A (MakeCo i arg ([] , \() ) ((ch1 , ch2)) ) body

  subst : (A : Type) {Cut : VarType} {_ : Check(canSubst (un▹ Cut) A)} -> (∀⇛ (Cut ⇒ A) ⊃ Cut ⁺ ⊃ A)
  subst A {D} {ch} Ψ body arg =  subst/anywhere A D {fst (check-andE ch)} {snd (check-andE{Wkn.canWeaken (un▹ D) A} ch)} i0 arg body 
  
