{-# OPTIONS --no-termination-check #-}

open import lib.Prelude 
open List using (_++_) 
open List.In using (_-_ ; skip)
open List.Subset
-- open List.InLists
-- open List.Subsets

open import foc.focbind.univ.examples.DefinedAtoms 
open import foc.focbind.univ.Types 
open import foc.focbind.univ.Semantics 
open import foc.focbind.univ.StrengthenT
open import foc.focbind.univ.ExchangeT
open import foc.focbind.univ.Map

module foc.focbind.univ.WeakenT where

module Weaken(S : SignatureInterface.Signature) where
  open Semantics(S)
  open SignatureInterface.Signature S
  open Map S

  module Ex = Exchange S
  module Str = Strengthen S

  mutual
    -- FIXME: convince of termination
    --        might be more efficient to share the visited list across recursive calls
    canWeaken* : List DefAtom -> DefAtom -> Type -> Bool
    canWeaken* seen D 1⁺ = True
    canWeaken* seen D 0⁺ = True
    canWeaken* seen D (A ⊗ B)  = (canWeaken* seen D A) andalso (canWeaken* seen D B)
    canWeaken* seen D (A ⊕ B)  = (canWeaken* seen D A) andalso (canWeaken* seen D B)
    canWeaken* seen D (list A) = canWeaken* seen D A
    canWeaken* seen D (D⁺ Dat) with List.In.find-in (\D0 -> DefinedAtoms.eq D0 Dat) seen
    ...                           | Some _ = True
    ...                           | None   = canWeakenD (Dat :: seen) D Dat
    canWeaken* seen D (Dat #)  = True
    canWeaken* seen D (Ψ ⇒* A) = canWeaken* seen D A
    canWeaken* seen D (□ A)  = True
    canWeaken* seen D (A ⊃ B)  = (Str.irrel D A) andalso (canWeaken* seen D B)
    canWeaken* seen D (∃c _) = False -- punt
    canWeaken* seen D (∀c _) = False
    canWeaken* seen D (∀⇒ A) = True -- because it's quantified!
    canWeaken* seen D (∀≤ Bound A) = DefinedAtoms.eq (un▹ Bound) D -- because it's (bounded) quantified
    canWeaken* seen D (∃⇒ A) = canWeaken* seen D A
    canWeaken* seen D ([ _ ]* _)  = True

    canWeakenD : List DefAtom -> DefAtom -> DefAtom -> Bool
    canWeakenD seen D Dat = List.all (\ A -> canWeaken* seen D A) (premises Dat)

  canWeaken : DefAtom -> Type -> Bool
  canWeaken = canWeaken* [] 

  open List.Subset

  mutual
    -- FIXME: termination
    -- computational induction on canWeaken* should work
    canWeaken*-mono : {Ds : List DefAtom} {Ds' : List DefAtom} -> (Ds ⊆ Ds')
                -> {D : DefAtom} -> (A : Type) -> Check (canWeaken* Ds D A) -> Check (canWeaken* Ds' D A)
    canWeaken*-mono w 1⁺ c = _
    canWeaken*-mono w 0⁺ _ = _
    canWeaken*-mono {seen} w {D} (A ⊗ B) c = 
      check-andI (canWeaken*-mono w A (fst (check-andE c))) 
                 (canWeaken*-mono w B (snd (check-andE{ canWeaken* seen D A } c))) 
    canWeaken*-mono {seen} w {D} (A ⊕ B) c = 
      check-andI (canWeaken*-mono w A (fst (check-andE c))) 
                 (canWeaken*-mono w B (snd (check-andE{ canWeaken* seen D A } c))) 
    canWeaken*-mono w (list A) c = canWeaken*-mono w A c
    canWeaken*-mono w (Ψ ⇒* A) c = canWeaken*-mono w A c
    canWeaken*-mono w (□ A) c = c
    canWeaken*-mono {seen} w {D} (A ⊃ B) c = 
      check-andI (fst (check-andE{ Str.irrel D A } c))
                 (canWeaken*-mono w B (snd (check-andE{ Str.irrel D A } c))) 
    canWeaken*-mono w (∃c _) () 
    canWeaken*-mono w (∀c _) () 
    canWeaken*-mono w (Dat #) c = c 
    canWeaken*-mono {seen} {seen'} w (D⁺ Dat) c with List.In.find-in (\D0 -> DefinedAtoms.eq D0 Dat) seen 
    ...                         | None with List.In.find-in (\D0 -> DefinedAtoms.eq D0 Dat) seen' 
    ...                                   | Some _ = _ 
    ...                                   | None =  canWeakenD-mono{Dat :: seen}{Dat :: seen'} (⊆-::-cong w) Dat c  
    canWeaken*-mono {seen} {seen'} w (D⁺ Dat) c | Some (_ , i , p) 
      with List.In.find-in (\D0 -> DefinedAtoms.eq D0 Dat) seen' | List.In.must-find (\D0 -> DefinedAtoms.eq D0 Dat) seen' (w i) p 
    ... | .(Some (x' , i' , p')) | (x' , i' , p' , Refl) = _ 
    canWeaken*-mono w (∃⇒ A) c = canWeaken*-mono w A c
    canWeaken*-mono w (∀⇒ A) c = _
    canWeaken*-mono w (∀≤ D A) c = c
    canWeaken*-mono w ([ _ ]* _) c = _

    canWeakenD-mono : {Ds : List DefAtom} {Ds' : List DefAtom} -> (Ds ⊆ Ds')
                -> {D : DefAtom} -> (Dat : DefAtom) -> Check (canWeakenD Ds D Dat) -> Check (canWeakenD Ds' D Dat)
    canWeakenD-mono {Ds} {Ds'} w {D} Dat c =  newall  where
      eworig = List.EW.from-checkall (premises Dat) c
      ewnew  = List.EW.mapew (\{A} c -> canWeaken*-mono w  A  c  ) eworig 
      newall = List.EW.to-checkall{Type}{(premises Dat)}{_} ewnew  

  Seen : DefAtom -> Set
  Seen D = Σ (\ (seen : List DefAtom) -> ({Dat : DefAtom} -> (Dat ∈ seen) -> Check(canWeakenD seen D Dat)))

  Co : Type -> Vars -> Vars -> Set
  Co A Ψ Ψ' = Σ \(D : VarType) -> 
                 Σ \(i : D ∈ Ψ') -> 
                   Σ \(seen : Seen (un▹ D)) -> Check(canWeaken* (fst seen) (un▹ D) A) × Id Ψ (Ψ' - i)

  Contra : Type -> Vars -> Vars -> Set
  Contra A Ψ Ψ' = Str.Co A Ψ Ψ'

  compat : AllNotstuckI Co Contra
  compat ((▹ Dat {c}) #) (D , i , seen , ch , Refl) = Inr (\ x -> (List.Subset.⊆remove i x ))
  compat (D⁺ Dat) (D , i , seen , ch , Refl) with List.In.find-in (\D0 -> DefinedAtoms.eq D0 Dat) (fst seen)
  ... | Some (Dat' , DatInSeeen , cheq) = Inr(
   (\{_} -> f) , 
   \ x -> ▹ (List.Subset.⊆remove i x ))  where
    f : {A : Type} (c : InΣ (Dat ⇐ A)) -> Co A _ _
    f {A} c = D , i , seen , (List.In.in-all (\ A -> canWeaken* (fst seen) (un▹ D ) A)
                                       ((snd seen)
                                        ((Id.subst (\ x -> x ∈ (fst seen)) 
                                                   (DefinedAtoms.certify-eq cheq) DatInSeeen) ) )
                                       {A} 
                                       (certify-premises c))  , Refl 
  ... | None = Inr ((\{_} -> f) , \ x -> ▹ (List.Subset.⊆remove i x ))  where
    f : {A : Type} (c : InΣ (Dat ⇐ A)) -> Co A _ _
    f {A} c = (D , i , 
               (Dat :: (fst seen) , 
                 \ {Dat'} i -> List.In.::-every {DefAtom}{Dat}
                               {\Dat'' -> Check(canWeakenD (Dat :: (fst seen)) (un▹ D) Dat'')}
                               {fst seen} 
                               (\ {Dat''} x ->   canWeakenD-mono iS  Dat'' ((snd seen) x)     ) ch  {Dat'} i ) , 
               (List.In.in-all (\ A -> canWeaken* (Dat :: (fst seen)) (un▹ D ) A)  
                         ch {A} ((certify-premises c)))  , 
               Refl)
  compat 1⁺ _ = Inr _ 
  compat 0⁺ _ = Inr _ 
  compat (A ⊗ B) (D , i , seen , ch , Refl) = Inr (
        ((D , i , seen ,  (fst (check-andE ch))  , Refl) ) ,
        (D , i , seen , (snd (check-andE{(canWeaken* (fst seen) (un▹ D) A)} ch)) , Refl))
  compat (A ⊕ B) (D , i , seen , ch , Refl) = Inr (
        ((D , i , seen ,  (fst (check-andE ch))  , Refl) ) ,
        (D , i , seen , (snd (check-andE{(canWeaken* (fst seen) (un▹ D) A)} ch)) , Refl))
  compat (□ A) _ = Inr _ 
  compat (list A) inv = Inr inv
  compat (∀c _) (_ , _ , _ , () , _)
  compat (∃c _) (_ , _ , _ , () , _)
  compat (Ψ ⇒* A) (D , i , seen , ch , Refl) = Inr (D , skip Ψ i , seen  , ch  , (List.In.remove/append Ψ i) )
  compat (A ⊃ B) (D , i , seen , ch , Refl) = Inr (
    (D ,  i  , ( ([] , \()) ) ,  (fst (check-andE{(Str.irrel (un▹ D) A)} ch))  ,  Refl) , 
    (D , i , seen , (snd (check-andE{(Str.irrel (un▹ D) A)} ch)) , Refl))
  compat (∀⇒ A) {Ψ' = Ψ'} (D , wkni , seen , ch , Refl) = 
    Inl (\ e Ψ0 -> Ex.exchange/anywhere A ((\{_} -> left{Ψ0}) , (\{_} -> right{Ψ0})) (e (D :: Ψ0))  ) where
     left : ∀ {Ψ0} -> ((D :: Ψ0) ++ (Ψ' - wkni)) ⊆ (Ψ0 ++ Ψ')
     left {Ψ0} i with List.In.splitappend (D :: Ψ0) _ i
     ...            | Inl i0 = skip Ψ0 wkni  
     ...            | Inl (iS i') = List.In.skip/right Ψ0 i' 
     ...            | Inr (i') = skip Ψ0 (List.Subset.⊆remove wkni i') 

     right : ∀ {Ψ0} -> (Ψ0 ++ Ψ') ⊆ ((D :: Ψ0) ++ (Ψ' - wkni))
     right {Ψ0} i with List.In.splitappend Ψ0 _ i 
     ... | Inl i' = List.In.skip/right (D :: Ψ0) (iS i') 
     ... | Inr i' with List.In.indeq wkni i'
     ...             | Inl Refl = i0 
     ...             | Inr newi' =  skip (D :: Ψ0) newi' 
  compat (∀≤ (▹ Bound) A) {Ψ' = Ψ'} (▹ D {chD} , wkni , seen , ch , Refl) = Inl 
    \ e Ψ0 inbounds -> 
        Ex.exchange/anywhere A ((\{_} -> left{Ψ0}) , (\{_} -> right{Ψ0})) 
          (e (▹ D {chD} :: Ψ0) 
             ( check-andI ch inbounds ) ) where
     -- FIXME: can't seem to abstract this out without getting 'failed to solve contraint' error
     left : ∀ {Ψ0} -> ((▹ D {chD} :: Ψ0) ++ (Ψ' - wkni)) ⊆ (Ψ0 ++ Ψ')
     left {Ψ0} i with List.In.splitappend (▹ D {chD} :: Ψ0) _ i
     ...            | Inl i0 = skip Ψ0 wkni  
     ...            | Inl (iS i') = List.In.skip/right Ψ0 i' 
     ...            | Inr (i') = skip Ψ0 (List.Subset.⊆remove wkni i') 

     right : ∀ {Ψ0} -> (Ψ0 ++ Ψ') ⊆ ((▹ D {chD} :: Ψ0) ++ (Ψ' - wkni))
     right {Ψ0} i with List.In.splitappend Ψ0 _ i 
     ... | Inl i' = List.In.skip/right (▹ D {chD} :: Ψ0) (iS i') 
     ... | Inr i' with List.In.indeq wkni i'
     ...             | Inl Refl = i0 
     ...             | Inr newi' =  skip (▹ D {chD} :: Ψ0) newi' 

     
  compat (∃⇒ A) (D , i , seen , ch , Refl) = Inr (\ Ψ -> (D , skip Ψ i , seen  , ch  , (List.In.remove/append Ψ i) ))
  compat ([ _ ]* _) _ = Inr _

  Spec+ = record { Co = Co ; Contra = Contra ; notstuck = \ A Ψ Ψ' -> compat A } 

  weaken/anywhere : (A : Type) {D : VarType} -> {_ : Check(canWeaken (un▹ D) A)} -> Str.WknFn A D
  weaken/anywhere A {D} {ch} i e = Map1+.map Spec+ Str.strengthen-map
                         A ((D , i , ([] , \ ()) , ch , Refl)) e

  weaken : (A : Type) {D : VarType} {_ : Check (canWeaken (un▹ D) A)} -> (∀⇛ A ⊃ ([ D ] ⇒* A))
  weaken A {D} {ch} Ψ t = weaken/anywhere A {_} {ch} i0 t 

  weaken* : (A : Type) (Ψ : Vars) {canw : Check (List.all (\ D -> (canWeaken (un▹ D) A)) Ψ)}
                  -> (∀⇛ A ⊃ (Ψ ⇒* A))
  weaken* A [] Ψ0 e = e 
  weaken* A (D :: Ψ) {ch} Ψ0 e = weaken/anywhere A {D} {hd} i0 ((weaken* A Ψ {tl} Ψ0 e) )  where
    hd = fst (check-andE ch) 
    tl = snd (check-andE{(canWeaken (un▹ D) A)} ch)  

  weaken*/bounded : (A : Type) (Ψ : Vars) {D' : VarType}
                  -> (AllEq Ψ D')
                  -> {canw : Check (canWeaken (un▹ D') A)}
                  -> (∀⇛ A ⊃ (Ψ ⇒* A))
  weaken*/bounded A Ψ {▹ D'} all {canw} = 
    weaken* A Ψ {List.In.every-to-checkall Ψ (\D -> (canWeaken (un▹ D) A)) f} where 
      f : {D : VarType} -> D ∈ Ψ -> Check (canWeaken (un▹ D) A)
      f {▹ D} i with List.In.in-all _ all i
      ...    | eq = Id.subst (\D''' -> Check (canWeaken D''' A)) ((DefinedAtoms.certify-eq eq) ) canw 
