\ignore{
\begin{code}
open import lib.Prelude 
open List using (_++_) 
open List.In using (_-_)
-- open List.InLists
-- open List.Subsets

open import foc.focbind.univ.examples.DefinedAtoms 
open import foc.focbind.univ.Types 
open import foc.focbind.univ.Semantics 
open import foc.focbind.univ.Structural
open import foc.focbind.univ.examples.Sig
open import foc.focbind.univ.examples.VariableManip using (plus)

module foc.focbind.univ.examples.Arith where
  open Sig
  open Semantics(theSig)
  open Structural(theSig)
\end{code}}

We define a signature for the arithmetic example mentioned above:

\noagda\begin{code}
  zero : InΣ (natA ⇐ 1⁺)
  succ : InΣ (natA ⇐ nat)
  
  num     : InΣ (arithA ⇐ nat)
  binop   : InΣ (arithA ⇐ arith ⊗ 
                           (nat ⊃ nat ⊃ nat) ⊗ 
                           arith)
  letbind : InΣ (arithA ⇐ arith ⊗ (arithC ⇒ arith))
\end{code}

Natural numbers are specified by zero and successor.  Arithmetic
expressions are given as a mixed datatype, with ⇒ used to represent the
body of the \ttt{letbind} and ⊃ used to represent primops: Next, we
define an evaluation function that reduces an expression to a number:

\begin{code}
  eval : ⊡ (arith ⊃ nat)
  eval (num · n)               = n
  eval (binop · (e1 , f , e2)) = f (eval e1) (eval e2)
  eval (letbind · (e1 , e2))   = eval (subst arith _ e2 e1)
  eval (▹ ()) 
\end{code}

Evaluation maps closed arithmetic expressions to natural numbers.
Constants evaluate to themselves; binops are evaluated by applying their
code to the values of the arguments; let-binding is evaluated by
applying the \ttt{letbind}'s body \ttt{e2} to the expression \ttt{e1},
performing substitution,\footnote{The \ttt{arith} argument to
\ttt{subst} is the type \ttt{A} in the \ttt{D ⇒ A} argument to
substitution; Agda's type reconstruction procedure requires this
annotation.  The underscore is the context argument instantiating the ∀⇛
in the type of \ttt{subst}.  The credentials for performing substitution
are marked as an implicit argument, so there is no evidence of it in the
call to \ttt{subst}.} and then evaluating the result. A simple variation
would be to evaluate \ttt{e1} first and then substitute its value into
\ttt{e2}.  The final clause covers the case for variables with a
refutation pattern: there are no variables in the empty context.

\paragraph{Comparison.}  This example provides a nice illustration of
the benefits of our approach: Substitution is provided ``for free'' by
the framework, which infers that it is permissible to substitute for
\ttt{arithC} variables in \ttt{arith}.  The type system enforces the
invariant that evaluation produces a closed natural number.  

It is not possible to define the type \ttt{arith} in
Twelf/Delphin/Beluga, as LF representations cannot use computational
functions.  One could program this example in FreshML, but it would be
necessary to implement substitution directly for \ttt{arith}, as FreshML
does not provide a generic substitution operation.

Agda checks that \ttt{eval}'s pattern matching is exhaustive.  However,
Agda is not able to verify the termination of this function, as it
recurs on a substitution-instance of one of the inputs.  Setting aside
the computational functions in \ttt{binop}, it would be possible to get
the call-by-value version of this code to pass Twelf's termination
checker, which recognizes certain substitution instances as smaller.  We
have not yet investigated how to explain this induction principle to
Agda.

\ignore{
\begin{code}
  teste : ⊡ arith
  teste = letbind · (num · (succ · (zero · _)) ,
                     binop · (▹ i0 , (\ x y -> y) , num · (zero · _)))

  teste2 : ⊡ arith
  teste2 = letbind · (num · (succ · (zero · _)) ,
                      binop · (▹ i0 , (weaken (nat ⊃ nat ⊃ nat) [] plus)  , num · (succ · (succ · (zero · _)))))

  -- 0
  teste3 : ⊡ arith
  teste3 = letbind · (num · (succ · (zero · _)) ,
                      letbind · (num · (zero · _),
                                 binop · (▹ i0 , (\ x y -> x) , ▹ (iS i0))))

  test = eval teste3
\end{code}}