\begin{figure}

\noagda\begin{code}
  napp : InΣ (neuA ⇐ neu ⊗ sem)
  neut : InΣ (semA ⇐ neu)
  slam : InΣ (semA ⇐ (∀≤ neuC (sem ⊃ sem)))
\end{code}

\hrule

\ignore{
\begin{code}
open import lib.Prelude 
open List using (_++_) 
open List.In using (_-_)
-- open List.InLists
-- open List.Subsets

open import foc.focbind.univ.examples.DefinedAtoms 
open import foc.focbind.univ.Types 
open import foc.focbind.univ.Semantics 
open import foc.focbind.univ.Structural
open import foc.focbind.univ.examples.Sig

module foc.focbind.univ.examples.NBE where
open Sig
open Semantics(theSig)
open Structural(theSig)

module Reify where 
 mutual
\end{code}}

\begin{code}
  reifyn : ∀⇛ ∀c \ Ψs -> (var2var neuC Ψs expC)  
         ⊃  [ Ψs ]* neu  ⊃  exp
  reifyn Ψe Ψs σ (▹ x) = ▹ (σ x)
  reifyn Ψe Ψs σ (napp · (n , s)) = 
    app · (reifyn Ψe Ψs σ n  , reify Ψe Ψs σ s) 

  reify :  ∀⇛ ∀c \ Ψs -> (var2var neuC Ψs expC) 
        ⊃  [ Ψs ]* sem  ⊃ exp
  reify Ψe Ψs σ (slam · φ) = 
    lam ·  reify (Ψe ,, expC) (Ψs ,, neuC) 
                 (extendv2v Ψs Ψe σ)
                 (φ [ neuC ] _ (neut · (▹ i0))) 
  reify Ψe Ψs σ (neut · n) = reifyn Ψe Ψs σ n 
  reify Ψe Ψs σ (▹ x) = impossible x
\end{code}

\hrule 

\ignore{
\begin{code}
module Eval where
\end{code}}

\begin{code}
  appsem : ∀⇛ sem ⊃ sem ⊃ sem
  appsem _ (slam · φ) s2 = φ [] _ s2
  appsem _ (neut · n) s2 = neut · (napp · (n , s2)) 
  appsem _ (▹ x) _       =  impossible x

  evalenv : Vars -> Type
  evalenv Ψs = (expC #) ⊃ ([ Ψs ]* sem)

  eval : ∀⇛ ∀c \Ψs -> evalenv Ψs  
       ⊃  exp  ⊃  ([ Ψs ]* sem)
  eval Ψe Ψs σ (▹ x) = σ x
  eval Ψe Ψs σ (app · (e1 , e2)) = 
       appsem Ψs (eval Ψe Ψs σ e1) (eval Ψe Ψs σ e2) 
  eval Ψe Ψs σ (lam · e) = slam · φ where
   φ : < Ψs > (∀≤ neuC (sem ⊃ sem))
   φ Ψ' bnd s' = eval (Ψe ,, expC) (Ψs + Ψ') σ' e where
      σ' : < Ψe > (expC ⇒ (evalenv (Ψs + Ψ')))
      σ' i0     = s' 
      σ' (iS i) = weaken*/bounded sem Ψ' bnd Ψs (σ i)

\end{code}
\caption{Normalization by evaluation}
\label{fig:nbe}
\end{figure}

In Figure~\ref{fig:nbe}, we present a serious example mixing binding and
computation, β-normalization-by-evaluation for the untyped λ-calculus.
NBE works by giving the syntax a semantics in terms of computational
functions (evaluation) and then reading back a normal form
(reification).  The NBE algorithm is similar to a Kripke logical
relations argument, where one defines a type- and context-indexed family
of relations $\sem{A} \:\dsd{in}\: \Psi$.  The key clause of this
definition is:

\[
(\sem{A \,\dsd{arrow}\, B} \:\dsd{in}\: \Psi) = \forall \Psi'.\: (\sem{A} \:\dsd{in}\: \Psi\mathord{,}\Psi') \supset (\sem{B} \:\dsd{in}\: \Psi\mathord{,}\Psi')
\]

That is, the meaning of $A \,\dsd{arrow}\, B$ in $\Psi$ is a function
that, for any future extension of the context, maps the meaning of $A$
in that extension to the meaning of $B$ in that extension.  In our type
theory, we represent (a simply-typed version of) this logical relation
as a datatype \dsd{sem}.  The datatype constructor corresponding to the
above clause would have the following type:
\noagda\begin{code}
sem ⇐ (∀⇒ sem ⊃ sem)
\end{code}
However, for the argument to go through, we must ensure that the context
extension Ψ' consists only of variables of a specific type \ttt{neu}, so we use 
a bounded context quantifier below.  

We represent the semantics by the datatypes \ttt{neu} and \ttt{sem} in
Figure~\ref{fig:nbe}.  The type \ttt{neu} (neutral terms) consists of
variables or neutral terms applied to semantic arguments (\ttt{napp});
these are the standard neutral proofs in natural deduction.  A \ttt{sem}
(semantic term) is either a neutral term or a semantic function.  A
semantic function of type \ttt{(∀≤ neuC (sem ⊃ sem))} is a computational
function that works in any extension of the context consisting entirely
of \ttt{neu} variables.  

We define reification first, via two mutually recursive functions,
\ttt{reifyn} (for neutral terms) and \ttt{reify} (for semantic terms).
It is typical in logical relations arguments to use two independent
contexts, one for the syntax and one for the semantics.  Thus, we
parametrize these functions by two contexts, one consisting for
\ttt{neu} variables for the semantics, and the other consisting of
\ttt{exp} variables for the syntax.  We will write \ttt{Ψs} for the
former and \ttt{Ψe} for the latter.  

In the type of \ttt{reify}, we must name one of these contexts, because
each context scopes over two disconnected parts of the type.  We choose
to name the semantic context and let the expression context be the
ambient one.  The outer \ttt{∀⇛} thus binds the expression context,
whereas we use the full binding form \ttt{∀c} for the semantic context.
The type of \ttt{reify} then says that, under some condition expressed
by the type \ttt{var2var}, \ttt{reify} maps semantics in the semantic
context (recall that \ttt{[ Ψ ]* A} stands for \ttt{□ (Ψ ⇒* A)}) to
expressions (in the ambient expression context).  The type \ttt{var2var
C1 Ψ1 C2} means that every variable of type \ttt{C1} in \ttt{Ψ1} maps to
a variable of type \ttt{C2} in the ambient context.  It is defined in a
library as follows:
\noagda\begin{code}
 var2var : VarType -> Vars -> VarType -> Type
 var2var C1 Ψ1 C2 = ([ Ψ1 ]* (C1 #)) ⊃ (C2 #)
\end{code}

Even though \ttt{reify} is given a precise type describing the scoping
of variables, its code is as simple as one could want.  To reify neutral
terms: The reification of a variable is the variable given in the
substitution.  The reification of an application is the application of
the reifications. To reify semantic terms: The reification of a function
\ttt{(slam · φ)} is the λ-abstraction of the reification of an instance
of φ. In the recursive call, the expression context is extended with a
new \ttt{exp} variable (which is bound by the \ttt{lam}) and the
semantic context is extended with a new \ttt{neu} variable.  We
instantiate the semantic function φ, which anticipates extensions of the
context, with this one-variable extension 
(\ttt{[ x ]} constructs a singleton list), and apply it to the variable.
The library function \ttt{extendv2v} makes the "parallel" extension of a
\ttt{var2var} in the obvious way, mapping the one new variable to the
other:
\noagda\begin{code}
 extendv2v : {D1 D2 : VarType} -> (Ψs : Vars) 
        -> ∀⇛ (var2var D1 Ψs D2)  
            ⊃  D2 ⇒ (var2var D1 (D1 :: Ψs) D2)
 extendv2v Ψs Ψe σ (i0)   = i0
 extendv2v Ψs Ψe σ (iS i) = iS (σ i)
\end{code}
The neutral-to-semantic coercion is reified recursively, and we disallow
\ttt{sem} variables from the context.

To define evaluation, we first define an auxiliary function \ttt{appsem}
that applies one semantic term to another.  This requires a
case-analysis of the function term: when it is an \ttt{slam} (i.e. the
application is a β-redex), we apply the embedded computational function,
choosing the nil context extension, and letting the argument be
\ttt{s2}.  When the function term is neutral, we make a longer neutral
term.

The type of \ttt{eval} is symmetric to \ttt{reify}, except the
environment that we carry along in the induction maps expression
variables to semantic \emph{terms} rather than just variables.  The type
\ttt{evalenv Ψs} means that every expression variable in the ambient
context is mapped to a semantic value in \ttt{Ψs}.  A variable is
evaluated by looking it up; an application is evaluated by combining the
recursive results with semantic application.  A \ttt{lam} is evaluated
to an \ttt{slam} whose body φ has the type indicated in the figure.
When given a context extension Ψ' and an argument \ttt{s'} in that
extension, φ evaluates the original body \ttt{e} in an extended
substitution.  The new substitution \ttt{σ'} maps the λ-bound variable
\ttt{i0} to the provided semantic value, and defers to σ on all other
variables.  However, σ provides values in \ttt{Ψs}, which must be
weakened into the extension Ψ'.  Fortunately, the bounded quantifier
provides sufficient evidence to show that weakening can be performed in
this case, because \ttt{sem}'s can be weakened with \ttt{neu} variables.

Normalization is defined by composing evaluation and reification.  We
define a normalizer for closed λ-terms as follows:

\ignore{
\begin{code}
module Norm where
  open Reify
  open Eval
\end{code}}

\begin{code}
  emptyv2v : ⊡ (var2var neuC [] expC)
  emptyv2v ()

  emptyevalenv : ⊡ (evalenv [])
  emptyevalenv ()

  norm : ⊡ (exp ⊃ exp)
  norm e = reify [] [] emptyv2v (eval [] [] emptyevalenv e) 
\end{code}
%
Our type system has verified the scope-correctness of this code, proving
that it maps closed terms to closed terms. Amusingly, Agda accepts the
termination of this evaluator for the untyped λ-calculus, provided that
we have told it to ignore its issues with our universe itself---a nice
illustration of the need for the positivity check on datatypes.

\paragraph{Comparison.}  The type \ttt{sem} is a truly mixed datatype:
the premise \ttt{(∀≤ neuC (sem ⊃ sem))} uses both ⇒ and ⊃ (recall that
there is a ⇒ buried in the definition of ∀≤).  Because it uses ⊃ in a
recursive datatype, it is not representable in LF.  Because it uses ⇒,
it would not even be representable in Delphin/Beluga extended with
standard recursive types (that did not interact with the LF part of the
language).  Despite the fact that our implementation enforces strong
invariants about the scope of variables, the code is essentially as
simple as the FreshML version described by
\citet{shinwell+03freshmlicfp}, aside from the need to pass the contexts
\ttt{Ψe} and \ttt{Ψs} along.  Invariants about variable scoping can be
proved in Pure FreshML~\citep{pottier07staticfreshml}, but we would like
to enforce these invariants within a type system, not using an external
specification logic.  Relative to a direct implementation in Agda, our
framework provides the weakening function needed in the final case of
\ttt{eval} for free.

\ignore{
\begin{code}
module Test where

  k : ⊡ exp
  k = (lam · (lam · ▹ (iS i0)))

  s : ⊡ exp
  s = (lam · (lam · (lam · (app · (app · (▹ (iS (iS i0)) , ▹ i0) , app · (▹ (iS i0) , ▹ i0))))))

  o : ⊡ exp
  o = lam · (app · (▹ i0 , ▹ i0))

  i : ⊡ exp
  i = lam · (▹ i0)

  i' : ⊡ exp
  i' = app · (app · (s , k) , s)

  -- should be o
  test : ⊡ exp
  test = Norm.norm (app · (app · (k , o) , i))

  -- should be i
  test' = Norm.norm i'

  loop : ⊡ exp
  loop = app · (o , o)

\end{code}}
