open import lib.Prelude 
open List using (_++_) 
open List.In using (_-_)
-- open List.InLists
-- open List.Subsets

open import foc.focbind.univ.examples.DefinedAtoms 
open import foc.focbind.univ.Types 
open import foc.focbind.univ.Semantics

module foc.focbind.univ.examples.Sig where

module Sig where
  open DefinedAtoms public
  open Types

  -- wrap the defined atoms as datatypes
  nat = D⁺ natA
  exp = D⁺ expA
  clos = D⁺ closA
  tp = D⁺ tpA 
  tm = D⁺ tmA 
  sem = D⁺ semA
  neu = D⁺ neuA
  sem2 = D⁺ sem2A
  neu2 = D⁺ neu2A
  sem3 = D⁺ sem3A
  neu3 = D⁺ neu3A
  arith = D⁺ arithA
  comb = D⁺ combA

  -- wrap the defined atoms with proofs they can go in the context
  expC = ▹ expA
  neuC = ▹ neuA
  neu2C = ▹ neu2A
  neu3C = ▹ neu3A
  tmC  = ▹ tmA
  arithC  = ▹ arithA
  combC  = ▹ combA

  data InΣ : Rule -> Set where
    -- nats
    zero : InΣ (natA ⇐ 1⁺)
    succ : InΣ (natA ⇐ nat)

    -- expressions
    lam  : InΣ (expA ⇐ expC ⇒ exp)
    app  : InΣ (expA ⇐ exp ⊗ exp)

    -- closures
    closure : InΣ (closA ⇐ (∃⇒ (expC ⇒ exp) ⊗ (expC #  ⊃  □ clos)))

    -- stlc 
    arr  : InΣ (tpA ⇐ tp ⊗ tp)
    prod : InΣ (tpA ⇐ tp ⊗ tp)
    unit : InΣ (tpA ⇐ 1⁺)

    fn   : InΣ (tmA ⇐ □ tp ⊗ (tmC ⇒ tm))
    ap   : InΣ (tmA ⇐ tm ⊗ tm)
    pair : InΣ (tmA ⇐ tm ⊗ tm)
    pi1  : InΣ (tmA ⇐ tm)
    pi2  : InΣ (tmA ⇐ tm)
    mt   : InΣ (tmA ⇐ 1⁺)

    -- sem and neu
    napp : InΣ (neuA ⇐ neu ⊗ sem)
    neut : InΣ (semA ⇐ neu)
    slam : InΣ (semA ⇐ (∀≤ neuC (sem ⊃ sem)))

    -- variations:
    napp3 : InΣ (neu3A ⇐ neu3 ⊗ sem3)
    neut3 : InΣ (sem3A ⇐ neu3)
    slam3 : InΣ (sem3A ⇐ (∀⇒ sem3 ⊃ sem3))
    napp2 : InΣ (neu2A ⇐ neu2 ⊗ sem2)
    neut2 : InΣ (sem2A ⇐ neu2)
    slam2 : InΣ (sem2A ⇐ (∀c \ Ψ -> (Ψ extendsfor neu2C) ⊃ [ Ψ ]* (sem2 ⊃ sem2)))

    -- arith 
    num     : InΣ (arithA ⇐ nat)
    binop   : InΣ (arithA ⇐ arith ⊗ (nat ⊃ nat ⊃ nat) ⊗ arith)
    letbind : InΣ (arithA ⇐ arith ⊗ (arithC ⇒ arith))

    -- comb 
    comb/s   : InΣ (combA ⇐ 1⁺)
    comb/k   : InΣ (combA ⇐ 1⁺)
    comb/i   : InΣ (combA ⇐ 1⁺)
    comb/app : InΣ (combA ⇐ comb ⊗ comb)

  -- FIXME: is there a way to not say things twice while
  -- (1) being able to automate the structural properties' conditions
  -- (2) pattern-matching with constants?

  typeof : ∀ {A D} -> InΣ (D ⇐ A) -> Type
  typeof {A} _ = A

  premises : DefinedAtoms.Atom -> List Type
  premises natA = typeof zero :: typeof succ :: []
  premises expA = typeof lam :: typeof app :: []
  premises closA = typeof closure :: []
  premises tpA = typeof arr :: typeof prod :: typeof unit :: []
  premises tmA = typeof fn :: typeof ap :: typeof pair :: typeof pi1 :: typeof pi2 :: typeof mt :: []
  premises neuA = typeof napp :: []
  premises semA = typeof neut :: typeof slam :: [] 
  premises neu2A = typeof napp2 :: []
  premises sem2A = typeof neut2 :: typeof slam2 :: [] 
  premises arithA = typeof num :: typeof binop :: typeof letbind :: []
  premises sem3A = typeof neut3 :: typeof slam3 :: []
  premises neu3A = typeof napp3 :: []
  premises combA = typeof comb/s :: typeof comb/k :: typeof comb/i :: typeof comb/app :: []
  premises ftpA = []

  certify-premises : {D : DefinedAtoms.Atom} {A : Types.Type} -> InΣ(Types._⇐_ D A) -> A ∈ (premises D)
  certify-premises zero = i0 
  certify-premises succ = iS i0
  certify-premises lam = i0
  certify-premises app = iS i0 
  certify-premises closure = i0 
  certify-premises arr = i0 
  certify-premises prod = iS i0 
  certify-premises unit = iS (iS i0) 
  certify-premises fn = i0 
  certify-premises ap = iS i0 
  certify-premises pair = iS (iS i0) 
  certify-premises pi1 = iS (iS (iS i0)) 
  certify-premises pi2 =  iS (iS (iS (iS i0))) 
  certify-premises mt = iS (iS (iS (iS (iS i0)))) 
  certify-premises napp = i0 
  certify-premises neut = i0 
  certify-premises slam = iS i0 
  certify-premises napp2 = i0
  certify-premises neut2 = i0 
  certify-premises slam2 = iS i0 
  certify-premises num = i0
  certify-premises binop = iS i0
  certify-premises letbind = iS (iS i0)
  certify-premises napp3 = i0 
  certify-premises neut3 = i0
  certify-premises slam3 = iS i0
  certify-premises comb/s = i0
  certify-premises comb/k = (iS i0)
  certify-premises comb/i = (iS (iS i0))
  certify-premises comb/app = (iS (iS (iS i0)))

  theSig : SignatureInterface.Signature
  theSig = record { InΣ = InΣ ; premises = premises  ; certify-premises = certify-premises }